"""Configuration for Garfield"""
import os
import seaborn as sns
import matplotlib as mpl

class GarfieldConfig:
    """configuration class for Garfield"""
    def __init__(self,
                 workdir='./result_Garfield',
                 n_jobs=1):
        self.workdir = workdir
        self.n_jobs = n_jobs
        self.set_gf_params()

    def set_figure_params(self,
                          context='notebook',
                          style='white',
                          palette='deep',
                          font='sans-serif',
                          font_scale=1.1,
                          color_codes=True,
                          dpi=80,
                          dpi_save=150,
                          fig_size=[5.4, 4.8],
                          rc=None):
        """ Set global parameters for figures. Modified from sns.set()

        Parameters
        ----------
        context : string or dict
            Plotting context parameters, see `seaborn.plotting_context`
        style: `string`,optional (default: 'white')
            Axes style parameters, see `seaborn.axes_style`
        palette : string or sequence
            Color palette, see `seaborn.color_palette`
        font_scale: `float`, optional (default: 1.3)
            Separate scaling factor to independently
            scale the size of the font elements.
        color_codes : `bool`, optional (default: True)
            If ``True`` and ``palette`` is a seaborn palette,
            remap the shorthand color codes (e.g. "b", "g", "r", etc.)
            to the colors from this palette.
        dpi: `int`,optional (default: 80)
            Resolution of rendered figures.
        dpi_save: `int`,optional (default: 150)
            Resolution of saved figures.
        rc: `dict`,optional (default: None)
            rc settings properties.
            Parameter mappings to override the values in the preset style.
            Please see "`matplotlibrc file
            <https://matplotlib.org/tutorials/introductory/customizing.html#a-sample-matplotlibrc-file>`__"
        """
        sns.set(context=context,
                style=style,
                palette=palette,
                font=font,
                font_scale=font_scale,
                color_codes=color_codes,
                rc={'figure.dpi': dpi,
                    'savefig.dpi': dpi_save,
                    'figure.figsize': fig_size,
                    'image.cmap': 'viridis',
                    'lines.markersize': 6,
                    'legend.columnspacing': 0.1,
                    'legend.borderaxespad': 0.1,
                    'legend.handletextpad': 0.1,
                    'pdf.fonttype': 42,
                    })
        if rc is not None:
            assert isinstance(rc, dict), "rc must be dict"
            for key, value in rc.items():
                if key in mpl.rcParams.keys():
                    mpl.rcParams[key] = value
                else:
                    raise Exception("unrecognized property '%s'" % key)

    def set_workdir(self, workdir=None):
        """Set working directory.

        Parameters
        ----------
        workdir: `str`, optional (default: None)
            Working directory.

        Returns
        -------
        """
        if(workdir is None):
            workdir = self.workdir
            print("Using default working directory.")
        if(not os.path.exists(workdir)):
            os.makedirs(workdir)
        self.workdir = workdir
        self.set_gf_params()
        print('Saving results in: %s' % workdir)

    def set_gf_params(self, config=None):
        """Set Garfield parameters

        Parameters
        ----------
        config : `dict`, optional
            Garfield training configuration parameters.

        Returns
        -------
        """
        default_config = {
            # Preprocessing options
            'adata_list': ['test.h5ad'],
            'profile': 'RNA',
            'data_type': None,
            'sub_data_type': None,
            'sample_col': None,
            'weight': 0.8,
            'graph_const_method': 'mu_std',
            'genome': None,
            'use_gene_weight': True,
            'use_top_pcs': False,
            'used_hvg': True,
            'min_features': 100,
            'min_cells': 3,
            'keep_mt': False,
            'target_sum': 1e4,
            'rna_n_top_features': 2000,
            'atac_n_top_features': 10000,
            'n_components': 50,
            'n_neighbors': 15,
            'metric': 'euclidean',
            'svd_solver': 'arpack',
            # datasets
            'used_pca_feat': False,
            'adj_key': 'spatial_connectivities',
            # data split parameters
            'edge_val_ratio': 0.1,
            'edge_test_ratio': 0.,
            'node_val_ratio': 0.1,
            'node_test_ratio': 0.,
            # model parameters
            'augment_type': 'svd',  # svd or dropout
            'svd_q': 5,  # if augment_type == 'svd'
            'use_FCencoder': True,
            'gnn_layer': 2,
            'conv_type': 'GAT',
            'hidden_dims': [128, 128],
            'bottle_neck_neurons': 20,
            'cluster_num': 20,
            'num_heads': 3,
            'dropout': 0.2,
            'concat': True,
            'drop_feature_rate': 0.2,
            'drop_edge_rate': 0.2,
            'used_edge_weight': True,
            'used_DSBN': False,
            'used_mmd': False,
            # data loader parameters
            'num_neighbors': 3,
            'loaders_n_hops': 3,
            'edge_batch_size': 256,
            'node_batch_size': None,
            # loss parameters
            'include_edge_recon_loss': True,
            'include_gene_expr_recon_loss': True,
            'lambda_latent_contrastive_instanceloss': 1.0,
            'lambda_latent_contrastive_clusterloss': 0.5,
            'lambda_gene_expr_recon': 300.,
            'lambda_edge_recon': 500000.,
            'lambda_omics_recon_mmd_loss': 0.2,
            # train parameters
            'n_epochs': 100,
            'n_epochs_no_edge_recon': 0,
            'learning_rate': 0.001,
            'weight_decay': 1e-05,
            'gradient_clipping': 5,
            # other parameters
            'latent_key': 'garfield_latent',
            'reload_best_model': True,
            'use_early_stopping': True,
            'early_stopping_kwargs': None,
            'monitor': True,
            'seed': 2024,
            'verbose': False,
            # 'loader_type': 'graphsaint',
        }

        ## user_config
        if config is None:
            config = {}  # 如果 `config` 为空则设为空字典

        user_config = config
        config = {**default_config, **user_config}

        assert config['adata_list'] is not None, 'The input `adata_list` must not be empty.'
        assert config['profile'] in ['RNA', 'ATAC', 'ADT', 'multi-modal', 'spatial'], \
            'The `profile` should be set as one of the `RNA`, `ATAC`, `ADT`, `multi-modal`, `spatial`.'
        if config['profile'] == 'multi-modal':
            assert config['data_type'] in ['Paired', 'UnPaired'], \
                'The `data_type` should be set as one of the `Paired`, `UnPaired`.'
            assert config['sub_data_type'] in [['rna', 'atac'], ['rna', 'adt']], \
                'The `sub_data_type` should be set as `["rna", "atac"]` or `["rna", "adt"]`.'
            if config['genome'] is not None:
                assert (config['genome'] in ['hg19', 'hg38', 'mm9', 'mm10']), \
                    "`genome` must be one of ['hg19','hg38','mm9','mm10']"
        if config['profile'] == 'spatial':
            assert config['data_type'] in ['single-modal', 'multi-modal'], \
                'The `data_type` should be set as one of the `single-modal`, `multi-modal`.'
        assert config['conv_type'] in ['GAT', 'GATv2Conv', 'GCN'], \
            'The current model only supports `GAT`, `GATv2Conv` or `GCN` layers.'

        self.gf_params = config


settings = GarfieldConfig()




from .decoders import (GATDecoder,
                       GCNDecoder)
from .encoders import (GATEncoder,
                       GCNEncoder)
from .utils import DSBatchNorm

__all__ = ["GATDecoder",
           "GCNDecoder",
           "GATEncoder",
           "GCNEncoder",
           "DSBatchNorm"]



from .metrics import eval_metrics, plot_eval_metrics
from .trainer import GarfieldTrainer
from .utils import EarlyStopping, print_progress

__all__ = ["eval_metrics",
           "plot_eval_metrics",
           "GarfieldTrainer",
           "EarlyStopping",
           "print_progress"]



"""
This module contains generic trainer functionalities, added as a Mixin to
the Trainer module.
"""
import inspect
import os
import dill
import pickle
from copy import deepcopy
from typing import Optional, Union

import numpy as np
import torch
import scanpy as sc
from anndata import AnnData, concat

class BaseMixin:
    """ Adapted from
        Title: scvi-tools
        Authors: Romain Lopez <romain_lopez@gmail.com>,
                 Adam Gayoso <adamgayoso@berkeley.edu>,
                 Galen Xing <gx2113@columbia.edu>
        Date: 14.12.2020
        Code version: 0.8.0-beta.0
        Availability: https://github.com/YosefLab/scvi-tools
        Link to the used code:
        https://github.com/YosefLab/scvi-tools/blob/0.8.0-beta.0/scvi/core/models/base.py
    """

    def _get_user_attributes(self):
        # returns all the self attributes defined in a model class, eg, self.is_trained_
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        attributes = [
            a for a in attributes if not (a[0].startswith("__") and a[0].endswith("__"))
        ]
        attributes = [a for a in attributes if not a[0].startswith("_abc_")]
        return attributes

    def _get_public_attributes(self):
        public_attributes = self._get_user_attributes()
        public_attributes = {a[0]: a[1] for a in public_attributes if a[0][-1] == "_"}
        return public_attributes

    def save(
            self,
            dir_path: str,
            overwrite: bool = False,
            save_anndata: bool = True,
            **anndata_write_kwargs,
    ):
        """Save the state of the model.
           Neither the trainer optimizer state nor the trainer history are saved.
           Parameters
           ----------
           dir_path
                Path to a directory.
           overwrite
                Overwrite existing data or not. If `False` and directory
                already exists at `dir_path`, error will be raised.
           save_anndata
                If True, also saves the anndata
           anndata_write_kwargs
                Kwargs for anndata write function
        """
        # get all the public attributes
        public_attributes = self._get_public_attributes()
        # save the model state dict and the trainer state dict only
        if not os.path.exists(dir_path) or overwrite:
            os.makedirs(dir_path, exist_ok=overwrite)
        else:
            raise ValueError(
                "{} already exists. Please provide an unexisting directory for saving.".format(
                    dir_path
                )
            )

        if save_anndata:
            self.adata.write(
                os.path.join(dir_path, "adata_ref.h5ad"), **anndata_write_kwargs
            )

        model_save_path = os.path.join(dir_path, "model_params.pt")
        attr_save_path = os.path.join(dir_path, "attr.pkl")
        varnames_save_path = os.path.join(dir_path, "var_names.csv")

        var_names = self.adata.var_names.astype(str)
        var_names = var_names.to_numpy()
        np.savetxt(varnames_save_path, var_names, fmt="%s")

        torch.save(self.model.state_dict(), model_save_path)

        # 现在使用 pickle.dump 来序列化，pickle 库已被替换为 dill
        with open(attr_save_path, "wb") as f:
            dill.dump(public_attributes, f)

    def _load_expand_params_from_dict(self, state_dict):
        load_state_dict = state_dict.copy()

        new_state_dict = self.model.state_dict()

        # 创建一个新的状态字典，仅包括在 new_state_dict 中也存在的键
        load_state_dict = {k: v for k, v in load_state_dict.items() if k in new_state_dict}

        for key, ten in new_state_dict.items():
            if key not in load_state_dict:
                load_state_dict[key] = ten

        self.model.load_state_dict(load_state_dict)

    @classmethod
    def _load_params(cls, dir_path: str, map_location: Optional[str] = None):
        setup_dict_path = os.path.join(dir_path, "attr.pkl")
        model_path = os.path.join(dir_path, "model_params.pt")
        varnames_path = os.path.join(dir_path, "var_names.csv")

        try:
            with open(setup_dict_path, "rb") as handle:
                attr_dict = dill.load(handle)
        # This catches the following error:
        # RuntimeError: Attempting to deserialize object on a CUDA device
        # but torch.cuda.is_available() is False.
        # If you are running on a CPU-only machine, please use torch.load with
        # map_location=torch.device('cpu') to map your storages to the CPU.
        except RuntimeError:
            with open(setup_dict_path, "rb") as handle:
                attr_dict = UnpicklerCpu(handle).load()

        model_state_dict = torch.load(model_path, map_location=map_location)

        var_names = np.genfromtxt(varnames_path, delimiter=",", dtype=str)

        return attr_dict, model_state_dict, var_names

class UnpicklerCpu(pickle.Unpickler):
    """Helps to pickle.load a model trained on GPU to CPU.

    See also https://github.com/pytorch/pytorch/issues/16797#issuecomment-633423219.
    """

    def find_class(self, module, name):
        if module == 'torch.storage' and name == '_load_from_bytes':
            return lambda b: torch.load(io.BytesIO(b), map_location='cpu')
        else:
            return super().find_class(module, name)



"""
This module contains the encoder used by the Garfield model.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.utils import dropout_adj
from torch_geometric.data import Data
from torch_geometric.nn import GCNConv, GATConv, GATv2Conv

from .utils import DSBatchNorm, drop_feature

class Projection(nn.Module):
    def __init__(self, in_dim : int, encoder_dim : int):
        super().__init__()
        self.layer = nn.Linear(in_dim, encoder_dim, bias=True)
        self.relu = nn.ReLU()
    def forward(self, x):
        return self.relu(self.layer(x))


class GATEncoder(nn.Module):
    def __init__(self, in_channels, hidden_dims, latent_dim, conv_type, use_FCencoder,
                 drop_feature_rate, drop_edge_rate, svd_q, num_heads, dropout, concat,
                 num_domains=1, used_edge_weight=False, used_DSBN=False):
        """
        Initializes the GATEncoder with multiple Graph Attention Network (GAT) layers, normalization layers,
        and optional fully connected (FC) encoder.

        Parameters
        ----------
        in_channels : int
            Number of input feature dimensions (length of each node's feature vector).
        hidden_dims : list[int]
            List of output dimensions for each hidden layer in the GAT.
        latent_dim : int
            Dimension of the latent feature representation produced by the encoder.
        conv_type : str
            Type of GAT convolution to use ('GAT' or 'GATv2Conv').
        use_FCencoder : bool
            Whether to use an additional fully connected encoder before the GAT layers.
        drop_feature_rate : float
            Dropout rate for node features during augmentation.
        drop_edge_rate : float
            Dropout rate for edges during augmentation.
        svd_q : int
            Rank for the low-rank SVD approximation used in augmentations. Default is 5.
        num_heads : int
            Number of attention heads for each GAT layer.
        dropout : float
            Dropout rate for GAT layers.
        concat : bool
            Whether to concatenate the outputs of all attention heads.
        num_domains : int, optional
            Number of domains for domain-specific batch normalization (DSBN). If `1`, regular batch normalization is used. Default is 1.
        used_edge_weight : bool, optional
            Whether to use edge weights in the GAT layers. Default is False.
        used_DSBN : bool, optional
            Whether to use domain-specific batch normalization (DSBN). Default is False.
        """
        super(GATEncoder, self).__init__()
        self.use_FCencoder = use_FCencoder
        self.drop_feature_rate = drop_feature_rate
        self.drop_edge_rate = drop_edge_rate
        self.svd_q = svd_q
        self.used_DSBN = used_DSBN
        self.used_edge_weight = used_edge_weight
        self.layers = nn.ModuleList()
        self.norm_layers = nn.ModuleList()

        # Choose GAT layer type based on `conv_type`
        GATLayer = GATConv if conv_type == 'GAT' else GATv2Conv

        # Initialize normalization layers based on `num_domains`
        for hidden_dim in hidden_dims:
            current_dim = hidden_dim * num_heads if concat else hidden_dim
            if num_domains == 1:
                norm_layer = nn.BatchNorm1d(current_dim)
            else:
                norm_layer = DSBatchNorm(current_dim, num_domains)
            self.norm_layers.append(norm_layer)

        # Initialize projection layer if `use_FCencoder` is True
        if self.use_FCencoder:
            encoder_dim = hidden_dims[0] * 2
            self.proj = Projection(in_channels, encoder_dim)
            current_dim = encoder_dim
        else:
            current_dim = in_channels

        # Initialize GAT layers
        for hidden_dim in hidden_dims:
            layer = GATLayer(
                in_channels=current_dim,
                out_channels=hidden_dim,
                heads=num_heads,
                dropout=dropout,
                concat=concat,
                edge_dim=1 if self.used_edge_weight else None
            )
            self.layers.append(layer)
            current_dim = hidden_dim * num_heads if concat else hidden_dim

        # Initialize the final mean and log standard deviation layers
        self.conv_mean = GATLayer(
            in_channels=current_dim,
            out_channels=latent_dim,
            heads=num_heads,
            concat=False,
            edge_dim=1 if self.used_edge_weight else None,
            dropout=dropout
        )
        self.conv_log_std = GATLayer(
            in_channels=current_dim,
            out_channels=latent_dim,
            heads=num_heads,
            concat=False,
            edge_dim=1 if self.used_edge_weight else None,
            dropout=dropout
        )

    def forward(self, data, decoder_type, augment_type):
        """
        Performs the forward pass through the GAT encoder, applying either omics or graph decoding, with optional augmentation.

        Parameters
        ----------
        data : Data
            PyTorch Geometric Data object containing node features, edge index, and other graph-related information.
        decoder_type : str
            Specifies the type of decoder to use, either 'omics' or 'graph'.
        augment_type : str, optional
            Specifies the type of augmentation to apply, either 'dropout' or 'svd'.
            If None, no augmentation is applied.

        Returns
        -------
        Tuple of torch.Tensor
            For 'omics' decoder type: (z_mean1, z_log_std1, z_mean2, z_log_std2), where:
            - z_mean1: Mean of the latent space from original input.
            - z_log_std1: Log standard deviation of the latent space from original input.
            - z_mean2: Mean of the latent space from augmented input.
            - z_log_std2: Log standard deviation of the latent space from augmented input.

            For 'graph' decoder type: (z_mean1, z_log_std1), where:
            - z_mean1: Mean of the latent space.
            - z_log_std1: Log standard deviation of the latent space.
        """
        x, edge_index_all, y = data.x, data.edge_index, data.y
        edge_index = edge_index_all[:, :2]

        if decoder_type == "omics":
            if augment_type is not None and augment_type == 'dropout':
                edge_weight = edge_index_all[:, 2] if self.used_edge_weight else None
                x_aug = drop_feature(x=x, drop_prob=self.drop_feature_rate)
                edge_index_aug = dropout_adj(edge_index, p=self.drop_edge_rate)[0]
                edge_weight_aug = edge_weight
            elif augment_type is not None and augment_type == 'svd':
                edge_weight = edge_index_all[:, 2].float()
                num_nodes = int(edge_index.max().item()) + 1
                sparse_adj = torch.sparse_coo_tensor(
                    indices=edge_index.t(),
                    values=edge_weight,
                    size=(num_nodes, num_nodes)
                )
                u, s, v = torch.svd_lowrank(sparse_adj, q=self.svd_q)
                recon_adj = (u @ torch.diag(s)) @ v.T
                sparse_adj = recon_adj.to_sparse()
                edge_index_aug = sparse_adj.indices()
                edge_weight_aug = sparse_adj.values().unsqueeze(1)
                x_aug = drop_feature(x=x, drop_prob=self.drop_feature_rate)
                # x_aug = data.x
            else:
                raise NotImplementedError(f"Unknown augment type: {augment_type}")

            if self.use_FCencoder:
                x = self.proj(x)
                x_aug = self.proj(x_aug)

            if not self.used_edge_weight and augment_type == 'svd':
                edge_weight = None
                edge_weight_aug = None
            z_mean1, z_log_std1 = self._forward_through_layers(x, edge_index, edge_weight, y)
            z_mean2, z_log_std2 = self._forward_through_layers(x_aug, edge_index_aug, edge_weight_aug, y)

            return z_mean1, z_log_std1, z_mean2, z_log_std2

        elif decoder_type == "graph":
            edge_weight = torch.ones(edge_index.shape[1]).unsqueeze(1).to(
                edge_index.device) if self.used_edge_weight else None

            if self.use_FCencoder:
                x = self.proj(x)

            z_mean1, z_log_std1 = self._forward_through_layers(x, edge_index, edge_weight, y)
            return z_mean1, z_log_std1

        else:
            raise NotImplementedError(f"Unknown decoder type: {decoder_type}")

    def _forward_through_layers(self, x, edge_index, edge_weight, y):
        """
        Helper function to pass the input features through multiple GAT layers and apply normalization.

        Parameters
        ----------
        x : torch.Tensor
            Node features (shape: [num_nodes, feature_dim]).
        edge_index : torch.Tensor
            Edge indices (shape: [2, num_edges]).
        edge_weight : torch.Tensor or None
            Edge weights for weighted GAT layers (optional, shape: [num_edges]).
        y : torch.Tensor
            Domain labels used for domain-specific batch normalization (DSBN) (shape: [num_nodes]).

        Returns
        -------
        Tuple of torch.Tensor
            - z_mean: Mean of the latent space after passing through the GAT layers (shape: [num_nodes, latent_dim]).
            - z_log_std: Log standard deviation of the latent space (shape: [num_nodes, latent_dim]).
        """
        for idx, layer in enumerate(self.layers):
            x, _ = layer(x, edge_index, edge_attr=edge_weight if self.used_edge_weight else None,
                         return_attention_weights=True)
            if self.used_DSBN and len(x) > 1:
                norm_layer = self.norm_layers[idx]
                if isinstance(norm_layer, DSBatchNorm):
                    x = norm_layer(x, y)
                else:
                    x = norm_layer(x)
            x = F.relu(x)

        z_mean, _ = self.conv_mean(x, edge_index, edge_attr=edge_weight if self.used_edge_weight else None,
                                   return_attention_weights=True)
        z_log_std, _ = self.conv_log_std(x, edge_index, edge_attr=edge_weight if self.used_edge_weight else None,
                                         return_attention_weights=True)

        return z_mean, z_log_std


### GCN encoder
class GCNEncoder(nn.Module):
    def __init__(self, in_channels, hidden_dims, latent_dim, use_FCencoder,
                 drop_feature_rate, drop_edge_rate, svd_q, dropout=0.2,
                 num_domains=1, used_edge_weight=False, used_DSBN=False):
        """
        Initializes the GCNEncoder with configurable options for feature projection, dropout, and domain-specific batch normalization (DSBN).

        Parameters
        ----------
        in_channels : int
            Number of input feature dimensions (length of each node's feature vector).
        hidden_dims : list[int]
            List of output dimensions for each hidden layer in the GCN.
        latent_dim : int
            Dimension of the latent feature representation produced by the encoder.
        use_FCencoder : bool
            Whether to use a fully connected encoder (FC encoder) before the GCN layers.
        drop_feature_rate : float
            Dropout rate for node features during augmentation.
        drop_edge_rate : float
            Dropout rate for edges during augmentation.
        svd_q : int
            Rank for the low-rank SVD approximation used in augmentations.
        dropout : float, optional
            Dropout rate applied to GCN layers, default is 0.2.
        num_domains : int, optional
            Number of domains for domain-specific batch normalization (DSBN). If `1`, regular batch normalization is used. Default is 1.
        used_edge_weight : bool, optional
            Whether to use edge weights in the GCN layers. Default is False.
        used_DSBN : bool, optional
            Whether to use domain-specific batch normalization (DSBN). Default is False.
        """
        super(GCNEncoder, self).__init__()

        self.use_FCencoder = use_FCencoder
        self.drop_feature_rate = drop_feature_rate
        self.drop_edge_rate = drop_edge_rate
        self.svd_q = svd_q
        self.used_DSBN = used_DSBN
        self.used_edge_weight = used_edge_weight
        self.norm = nn.ModuleList()

        # Apply feature projection if specified
        if self.use_FCencoder:
            encoder_dim = hidden_dims[0] * 2
            self.proj = Projection(in_channels, encoder_dim)
            current_dim = encoder_dim
        else:
            current_dim = in_channels

        # Initialize GCN layers
        self.gcn_layers = nn.ModuleList()
        total_layers = [current_dim] + hidden_dims
        for i in range(len(total_layers) - 1):
            self.gcn_layers.append(GCNConv(total_layers[i], total_layers[i + 1], dropout=dropout))

        # Initialize mean and log standard deviation layers
        self.gcn_mu = GCNConv(hidden_dims[-1], latent_dim, dropout=dropout)
        self.gcn_logvar = GCNConv(hidden_dims[-1], latent_dim, dropout=dropout)

        # Initialize normalization layers
        for current_dim in hidden_dims:
            if num_domains == 1:
                norm = nn.BatchNorm1d(current_dim)
            else:
                norm = DSBatchNorm(current_dim, num_domains)
            self.norm.append(norm)

    def forward(self, data, decoder_type, augment_type=None):
        """
        Forward pass through the GCN encoder, with optional augmentations such as dropout or SVD, and multiple decoding options.

        Parameters
        ----------
        data : Data
            PyTorch Geometric Data object containing node features, edge index, and other graph-related information.
        decoder_type : str
            Specifies the type of decoder to use, either 'omics' for gene expression data or 'graph' for graph structure.
        augment_type : str, optional
            Specifies the type of augmentation to apply, either 'dropout' or 'svd'. If None, no augmentation is applied.

        Returns
        -------
        Tuple of torch.Tensor
            For 'omics' decoder type: (z_mean1, z_log_std1, z_mean2, z_log_std2), where:
            - z_mean1: Mean of the latent space from original input.
            - z_log_std1: Log standard deviation of the latent space from original input.
            - z_mean2: Mean of the latent space from augmented input.
            - z_log_std2: Log standard deviation of the latent space from augmented input.

            For 'graph' decoder type: (z_mean1, z_log_std1), where:
            - z_mean1: Mean of the latent space.
            - z_log_std1: Log standard deviation of the latent space.
        """
        x, edge_index_all, y = data.x, data.edge_index, data.y
        edge_index = edge_index_all[:, :2]

        if decoder_type == "omics":
            if augment_type is not None and augment_type == 'dropout':
                edge_weight = edge_index_all[:, 2] if self.used_edge_weight else None
                x_aug = drop_feature(x=x, drop_prob=self.drop_feature_rate)
                edge_index_aug = dropout_adj(edge_index, p=self.drop_edge_rate)[0]
                edge_weight_aug = edge_weight
            elif augment_type is not None and augment_type == 'svd':
                edge_weight = edge_index_all[:, 2].float()
                num_nodes = int(edge_index.max().item()) + 1
                sparse_adj = torch.sparse_coo_tensor(
                    indices=edge_index.t(),
                    values=edge_weight,
                    size=(num_nodes, num_nodes)
                )
                u, s, v = torch.svd_lowrank(sparse_adj, q=self.svd_q)
                recon_adj = (u @ torch.diag(s)) @ v.T
                sparse_adj = recon_adj.to_sparse()
                edge_index_aug = sparse_adj.indices()
                edge_weight_aug = sparse_adj.values().unsqueeze(1)
                # x_aug = drop_feature(x=x, drop_prob=self.drop_feature_rate)
                x_aug = data.x
            else:
                raise NotImplementedError(f"Unknown augment type: {augment_type}")

            if self.use_FCencoder:
                x = self.proj(x)
                x_aug = self.proj(x_aug)

            if not self.used_edge_weight and augment_type == 'svd':
                edge_weight = None
                edge_weight_aug = None
            z_mean1, z_log_std1 = self._forward_through_layers(x, edge_index, edge_weight, y)
            z_mean2, z_log_std2 = self._forward_through_layers(x_aug, edge_index_aug, edge_weight_aug, y)

            return z_mean1, z_log_std1, z_mean2, z_log_std2

        elif decoder_type == "graph":
            edge_weight = torch.ones(edge_index.shape[1]).unsqueeze(1).to(
                edge_index.device) if self.used_edge_weight else None

            if self.use_FCencoder:
                x = self.proj(x)

            z_mean1, z_log_std1 = self._forward_through_layers(x, edge_index, edge_weight, y)
            return z_mean1, z_log_std1

        else:
            raise NotImplementedError(f"Unknown decoder type: {decoder_type}")

    def _forward_through_layers(self, x, edge_index, edge_weight, y):
        """
        Helper function that passes the node features through GCN layers and applies normalization.

        Parameters
        ----------
        x : torch.Tensor
            Node features (shape: [num_nodes, feature_dim]).
        edge_index : torch.Tensor
            Edge indices (shape: [2, num_edges]).
        edge_weight : torch.Tensor or None
            Edge weights for weighted GCN layers (optional, shape: [num_edges]).
        y : torch.Tensor
            Domain labels used for domain-specific batch normalization (DSBN) (shape: [num_nodes]).

        Returns
        -------
        Tuple of torch.Tensor
            - z_mean: Mean of the latent space after passing through the GCN layers (shape: [num_nodes, latent_dim]).
            - z_log_std: Log standard deviation of the latent space (shape: [num_nodes, latent_dim]).
        """
        for idx, layer in enumerate(self.gcn_layers):
            x = layer(x, edge_index, edge_weight=edge_weight)
            x = self._apply_normalization(x, y, idx)
            x = F.relu(x)

        z_mean = self.gcn_mu(x, edge_index, edge_weight=edge_weight)
        z_log_std = self.gcn_logvar(x, edge_index, edge_weight=edge_weight)
        return z_mean, z_log_std

    def _apply_normalization(self, x, y, idx):
        """
        Applies batch normalization or domain-specific batch normalization (DSBN) based on the model's configuration.

        Parameters
        ----------
        x : torch.Tensor
            Node features (shape: [num_nodes, feature_dim]).
        y : torch.Tensor
            Domain labels for DSBN (shape: [num_nodes]).
        idx : int
            Index of the current layer for normalization.

        Returns
        -------
        torch.Tensor
            Normalized node features after applying either batch normalization or DSBN (shape: [num_nodes, feature_dim]).
        """
        if self.used_DSBN and len(x) > 1:
            norm_layer = self.norm[idx]
            if isinstance(norm_layer, DSBatchNorm):
                x = norm_layer(x, y)
            else:
                x = norm_layer(x)
        return x


